import sys

from .__command__ import __start__, __add_server__, __del_server__, __show_configs__, __create_snapshot__, \
    __preview_snapshot__, __apply_snapshot__, __delete_snapshot__, __list_snapshot__, set_command
from .config import load_config
from .utils import out_print


def script_main():
    def print_version():
        import pkg_resources
        out_print(f'salesea {pkg_resources.get_distribution("salesea").version}')

    # 命令行解析
    import argparse
    parser = argparse.ArgumentParser(
        prog='salesea',
        usage='%(prog)s [command] [options]',
        description='This is an Nginx log collection tool.',
        add_help=False
    )
    # command
    parser.add_argument(
        'commands', nargs='*',
        help='Commands: add, rm, start, preview, backup, restore, list'
    )
    # options
    parser.add_argument(
        '-V', '--version', action='store_true',
        help='打印版本信息并退出'
    )
    parser.add_argument(
        '-h', '--help', action='store_true',
        help='显示帮助信息并退出'
    )
    parser.add_argument(
        '-c', '--config', type=str, default='salesea.row',
        help='指定配置文件'
    )
    # 不在帮助信息中显示
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help=argparse.SUPPRESS
    )
    parser.add_argument(
        '-k', '--apikey', type=str,
        help=argparse.SUPPRESS
    )
    args = parser.parse_args()

    if args.version:
        print_version()
        sys.exit()

    if args.commands:
        load_config(args.config)
        # 二级命令
        command, *data = args.commands

        if command not in ('start', 'add', 'rm', 'configs', 'backup', 'preview', 'restore', 'delete', 'list'):
            out_print(f'未知命令: {command}')
            sys.exit()

        set_command(command)

        if command == 'start':
            __start__(args.commands)

        if command == 'add':
            __add_server__(args.commands)

        if command == 'rm':
            __del_server__(args.commands)

        if command == 'configs':
            __show_configs__(args.commands)

        if command == 'backup':
            __create_snapshot__(args.commands)

        if command == 'preview':
            __preview_snapshot__(args.commands)

        if command == 'restore':
            __apply_snapshot__(args.commands)

        if command == 'delete':
            __delete_snapshot__(args.commands)

        if command == 'list':
            __list_snapshot__(args.commands)

    parser.print_help()


'''
    # 添加域名
    parser.add_argument(
        '-a', '--add', type=str,
        help='Add a domain name'
    )
    # 删除域名
    # parser.add_argument(
    #     '-r', '--remove', type=str,
    #     help='Remove a domain name'
    # )
    parser.add_argument(
        '-s', '--start', action='store_true',
        help='Start the service'
    )
    # 预览快照
    parser.add_argument(
        '-p', '--preview', type=str,
        help='Preview the snapshot'
    )
    # 创建快照
    parser.add_argument(
        '-b', '--backup', action='store_true',
        help='Create a snapshot'
    )
    # 恢复快照 输入快照ID
    parser.add_argument(
        '-r', '--restore', type=str,
        help='Restore the snapshot'
    )
    # 查看快照
    parser.add_argument(
        '-l', '--list', action='store_true',
        help='List snapshots'
    )
    # 删除快照 输入快照ID
    parser.add_argument(
        '-x', '--delete', type=str,
        help='Delete the snapshot'
    )

    args = parser.parse_args()


    if args.add:
        from .config import add_domain
        server_name = add_domain(args.add, args.apikey)
        out_print(f'域名已添加: {server_name}')
        sys.exit()

    # if args.remove:
    #     from .config import remove_domain
    #     result = remove_domain(args.remove)
    #     if result:
    #         out_print(f'域名已删除: {args.remove}')
    #     else:
    #         out_print(f'域名不存在: {args.remove}')
    #     sys.exit()



    from .config import NGINX_PATH
    from .nginx import Nginx
    nginx = Nginx(Path(NGINX_PATH) if NGINX_PATH else None)

    if args.preview:
        out_print('---------- 预览快照 ----------')
        configs = nginx.get_snapshot(args.preview)
        for filename, content in configs.items():
            out_print(f'# configuration file: {filename}')
            out_print(content)
        sys.exit()

    if args.backup:
        out_print('---------- 创建快照 ----------')
        nginx.save_snapshot()
        sys.exit()

    if args.restore:
        out_print('---------- 恢复快照 ----------')
        if confirm(f'请确认是否恢复快照[{args.restore}]?'):
            nginx.restore_snapshot(args.restore)
        sys.exit()

    if args.list:
        snapshots = nginx.get_snapshot_list()
        out_print('---------- 快照列表 ----------')
        if not snapshots:
            out_print('          [没有快照]')
            out_print('请使用 [salesea -b] 创建快照')
            sys.exit()
        for snapshot in snapshots:
            out_print(snapshot.get('name'))
        sys.exit()

    if args.delete:
        # 确认删除
        out_print('---------- 删除快照 ----------')
        if confirm(f'请确认是否删除快照[{args.delete}]?'):
            nginx.delete_snapshot(args.delete)
        sys.exit()

    parser.print_help()
    sys.exit()'''


def main():
    try:
        script_main()
    except KeyboardInterrupt:
        out_print("\n\nCtrl+C")



from .utils import out_print


# 出现 未解析到配置 错误时，打印解决方案
def print_solution():
    out_print('未解析到指定的Nginx服务器，可能原因：')
    out_print('    1. 您填写的[server_name]参数未能匹配到对应的Nginx服务器;')
    out_print('    2. Nginx服务器配置中[server_name]参数不能为空;')
    # out_print('    3. Nginx服务器未配置以下参数:')
    # out_print(
    #     '''
    #     ...
    #     html {
    #         ...
    #         log_format salesea '$remote_addr - $remote_user [$time_local] "$request" '
    #                          '$status $body_bytes_sent "$http_referer" '
    #                          '"$http_user_agent" "$http_x_forwarded_for"';
    #         ...
    #         server {
    #             ...
    #             access_log /var/log/nginx/access.log salesea;
    #             ...
    #         }
    #     }
    #     ...'''
    # )



import configparser
from pathlib import Path

from .log import logger
from .utils import out_print

#############################################################################
#####Config Parser###########################################################
LOG_FORMAT = '$remote_addr - $remote_user [$time_local] "$request" $status $body_bytes_sent "$http_referer" "$http_user_agent" "$http_x_forwarded_for"'
SERVER_MAP = {}
NGINX_PATH = None
CHECK_INTERVAL = 60
DEFAULT_CONF_FILE = "salesea.row"
CURRENT_CONF_FILE = DEFAULT_CONF_FILE


def load_config(CONF_FILE=DEFAULT_CONF_FILE):
    global SERVER_MAP, CURRENT_CONF_FILE
    CONF_FILE = Path(CONF_FILE)
    if not CONF_FILE.exists():
        logger.debug(f"[{CONF_FILE}] 配置文件不存在")
        CONF_FILE.touch()

    with CONF_FILE.open(mode='r', encoding='utf-8') as fp:
        while True:
            line = fp.readline()
            try:
                if line == '':
                    break
                SERVER_MAP[line.split()[0]] = line.split()[1]
            except Exception as e:
                logger.debug(f"配置文件格式错误: {line}")

    # if not SERVER_MAP:
    #     logger.info(f"无效配置: {CONF_FILE}")
    #     exit(1)
    CURRENT_CONF_FILE = CONF_FILE


def add_server(server_name=None, visit_apikey=None):
    # 询问用户
    while not server_name:
        server_name = input('请输入域名(必填): ')

    while not visit_apikey:
        visit_apikey = input('请输入访问密钥(必填): ')

    SERVER_MAP[server_name] = visit_apikey
    with CURRENT_CONF_FILE.open(mode='w', encoding='utf-8') as fp:
        for k, v in SERVER_MAP.items():
            fp.write(f"{k} {v}\n")

    return server_name


def remove_server(server_name=None):
    # 询问用户
    while not server_name:
        server_name = input('请输入域名(必填): ')

    if server_name in SERVER_MAP:
        SERVER_MAP.pop(server_name)
        with CURRENT_CONF_FILE.open(mode='w', encoding='utf-8') as fp:
            for k, v in SERVER_MAP.items():
                fp.write(f"{k} {v}\n")
        return True
    else:
        return False


def list_server():
    return SERVER_MAP


def generate_config():
    config = configparser.ConfigParser()
    config['nginx'] = {
        'server_name': '',
        'nginx_path': '',
    }
    config['request'] = {
        'concurrency': 10,
    }
    config['salesea'] = {
        'visit_apikey': '',
        'interval': 60,
    }
    # 询问用户
    while True:
        config['nginx']['server_name'] = input('请输入域名(必填): ')
        if config['nginx']['server_name']:
            break

    while True:
        config['salesea']['visit_apikey'] = input('请输入访问密钥(必填): ')
        if config['salesea']['visit_apikey']:
            break

    # 请输入日志扫描间隔 不能小于5秒
    while True:
        interval = config['salesea']['interval']
        try:
            i = input('请输入日志扫描间隔(默认60秒): ')
            interval = int(i) if i else 60
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except:
            continue

        if interval >= 5:
            config['salesea']['interval'] = str(interval)
            break
        out_print('日志扫描间隔不能小于5秒')

    with open(DEFAULT_CONF_FILE, 'w') as f:
        config.write(f)
        out_print(f'配置文件已生成: {DEFAULT_CONF_FILE}')
        exit(0)



# -*- coding: utf-8 -*-
# 解析Nginx日志, 根据开始时间进行筛选
from time import sleep

import requests
import json
import re
from pathlib import Path

from .nginx import Nginx, NginxException
from .log import logger
from .solution import print_solution


# 输入迭代器和一个数字n，返回n个元素的列表，直到迭代器耗尽，不使用islice， 使用yield
def take(iterable, n):
    result = []
    for i in iterable:
        result.append(i)
        if len(result) == n:
            yield result
            result = []
    if result:
        yield result


def launch():
    # 将PID写入文件
    # pid = os.getpid()
    # with open("salesea.pid", "w") as f:sys
    #     f.write(str(pid))
    #############################################################################
    #####Global Variable#########################################################
    datetime_format = "%Y-%m-%d %H:%M:%S %z"  # 时间格式
    nginx = None

    #############################################################################
    #####Get Access Servers######################################################

    def get_access_servers(server_list):
        nginx_path = nginx.nginx_path
        servers = []
        logger.debug(f"nginx_path: {nginx_path}")
        for server in nginx.servers:
            match = server.is_salesea() and server.eq_name(server_list) and server.eq_listen("443")
            logger.debug(
                f"server_name: [{'√' if match else '×'}]{' '.join([_ for _ in server.names])} => {server.listens}")

            if match:
                servers.append(server)
        return servers

    #############################################################################
    #####Parse Nginx Log########################################################
    def parse_nginx_log(logfile: Path, pattern):
        if not logfile.exists():
            logger.error(f"日志文件不存在: {str(logfile)}")
        else:
            offset = 0
            with open(logfile, mode="r+", encoding="utf-8") as fobj:
                while True:
                    line = fobj.readline().strip()
                    if line:
                        try:
                            row = re.match(pattern, line)
                            data = row.groupdict()
                            yield data
                        except Exception as e:
                            logger.error(f"解析日志错误: {e}")
                            logger.info(f"解析日志错误: {line}")
                    else:
                        # 获取当前偏移量
                        offset = fobj.tell()
                        # 定位文件指针
                        fobj.seek(offset)
                        # 清除0到偏移量之间的内容
                        fobj.truncate(0)
                        break

    #############################################################################
    #####Scheduler##############################################################
    def scheduler(interval):
        def decorator(func):
            def wrapper(*args, **kwargs):
                while True:
                    func(*args, **kwargs)
                    sleep(interval)

            return wrapper

        return decorator

    #############################################################################
    #####Utils###################################################################
    # 迭代器切割
    def chunked(iterable, n):
        for i in range(0, len(iterable), n):
            yield iterable[i:i + n]

    #############################################################################
    #####Main###################################################################
    def main():
        logfile = "./access.log"
        # 替换所有的$符号为(.*?)，并且将$符号后面的字符串作为分组名
        log_pattern = re.sub(r"\$[a-zA-Z_]\w*", r"(?P<\g<0>>.*?)", LOG_FORMAT)
        log_pattern = log_pattern.replace('$', '').replace('[', '\[').replace(']', '\]')
        servers = get_access_servers(SERVER_LIST)

        if servers:
            [logger.info(f'解析到配置：{server.names}:{server.listens} >> {server.logfile}') for server in servers]
        else:
            logger.error('未解析到指定的Nginx服务器\n')
            print_solution()
            exit(1)

        @scheduler(CHECK_INTERVAL)
        def task():
            count = 0
            with requests.session() as sess:
                for server in servers:
                    # logger.debug(f"开始扫描日志文件：{[str(server.logfile) for p in servers]}")
                    nginx_log_iter = parse_nginx_log(server.logfile, log_pattern)

                    for datas in take(nginx_log_iter, 100):
                        send_data = []
                        for data in datas:
                            try:
                                d = re.match(r"[a-zA-Z]+\s(?P<path>/.*?)(?P<query>\?.*?)?\s",
                                             data['request']).groupdict()
                                send_data.append({
                                    'visitApiKey': SERVER_MAP.get(server.default_name),
                                    'domain': server.default_name or server.names[0],
                                    'path': d.get('path'),
                                    'query': d.get('query'),
                                    'referrer': data.get('http_referer'),
                                    'user_agent': data.get('http_user_agent'),
                                    'ip': data.get('remote_addr'),
                                })
                            except Exception as e:
                                logger.error(f"解析日志错误: {e} - {data}")
                        try:
                            result: requests.Response = sess.post('https://dashboard.salesea.cn/api/visit/batch',
                                                                  data=json.dumps(send_data),
                                                                  headers={
                                                                      'Content-Type': 'application/json'
                                                                  })
                            count += len(send_data)
                            obj = result.json()
                            if obj.get('code') == 'error':
                                logger.error(f"响应错误: {obj} - {send_data}")
                            else:
                                logger.info(f"响应结果: {obj}")
                        except Exception as e:
                            logger.error(f"请求错误: {e} - {send_data}")

            logger.__getattribute__('info' if count else 'debug')(f"解析到[{count}]条日志")

        task()

    try:
        from .config import NGINX_PATH, SERVER_MAP, CHECK_INTERVAL, LOG_FORMAT

        SERVER_LIST = tuple(SERVER_MAP.keys())

        try:
            nginx = Nginx(Path(NGINX_PATH) if NGINX_PATH else None)

            #############################################################################
            #####Print Config############################################################
            logger.debug(f"#" * 65)
            logger.debug(f"# nginx_path: {nginx.nginx_path}")
            logger.debug(f"# server_name: {SERVER_LIST}")
            logger.debug(f"# check_interval: {CHECK_INTERVAL}")
            logger.debug(f"# visit_apikey: {'*' * 8} ({len(tuple(SERVER_MAP.values()) or '')})")
            logger.debug(f"#" * 65)

            logger.debug(f'解析到[http]配置文件[{nginx.httpConfigPath}]')
            for server in nginx.servers:
                logger.debug(f'解析到[server]配置文件[{server.serverConfigPath}]')

            nginx.create_snapshot()
            nginx.write_salesea_log_format()
            nginx.include_saleseas(SERVER_MAP)
        except NginxException as e:
            logger.error(f"{e}")
            exit(1)

        main()
    except KeyboardInterrupt:
        logger.info("用户终止程序")



import logging
import os
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

#############################################################################
#####Logger Handler##########################################################
# linux 和 windows 日志存储路径
# linux: /var/log/salesea/salesea.log
# windows: C:\Users\{username}\AppData\Local\salesea\logs\salesea.log

if os.name == "posix":
    log_path = Path("/var/log/salesea/")
else:
    log_path = Path(os.path.expandvars(r"%LOCALAPPDATA%")) / "salesea" / "logs"

log_path.mkdir(parents=True, exist_ok=True)

logger = logging.getLogger(__name__)  # 日志对象
logger.level = logging.INFO  # 日志级别
handler = logging.StreamHandler()  # 日志处理器
handler.formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")  # 日志格式
logger.addHandler(handler)  # 添加日志处理器
# 日志文件处理器
fh = logging.handlers.RotatingFileHandler(log_path / "salesea.log", maxBytes=256 * 1024 * 10, backupCount=10)
fh.formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")  # 日志格式
logger.addHandler(fh)  # 添加日志文件处理器



import argparse
import sys
from pathlib import Path

from .log import logger
from .utils import out_print

parser = argparse.ArgumentParser(
    prog='salesea',
    usage='%(prog)s [options]',
    description='This is an Nginx log collection tool.',
    add_help=False
)

parser.add_argument(
    '-c', '--config', type=str, help=argparse.SUPPRESS
)


def set_command(command):
    parser.usage = f'%(prog)s {command} [options]'


# 命令行确认
def __confirm__(prompt: str, default: bool = False) -> bool:
    """Confirm with user input"""
    if default:
        prompt = f'{prompt} (Y/n) '
    else:
        prompt = f'{prompt} (y/N) '
    while True:
        choice = input(prompt).lower()
        if choice in ('y', 'yes'):
            return True
        elif choice in ('n', 'no'):
            return False
        elif choice == '':
            return default
        else:
            print('请回答 yes 或 no.')


def __start__(argv):
    parser.add_argument('start', nargs='?', help='启动程序')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')
    parser.add_argument('-d', '--debug', action='store_true', help='调试模式')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    if args.debug:
        logger.setLevel('DEBUG')

    from . import launch
    launch()
    sys.exit()


def __add_server__(argv):
    parser.add_argument('add', nargs='?', help='添加配置')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')
    parser.add_argument('domain', nargs='?', help='指定域名')
    parser.add_argument('-k', '--apikey', type=str, help='指定API Key')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    from .config import add_server
    server_name = add_server(args.domain, args.apikey)
    out_print(f'域名已添加: {server_name}')
    sys.exit()


def __del_server__(argv):
    parser.add_argument('rm', nargs='?', help='删除配置')
    parser.add_argument('domain', nargs='?', help='列出快照')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    from .config import remove_server
    if not args.datas:
        out_print(f'未指定域名')
        sys.exit()

    if __confirm__(f'确定要删除域名[{args.domain}]吗？'):
        result = remove_server(args.domain)
        out_print(f'域名[{args.domain}]已删除' if result else f'域名[{args.domain}]不存在')

    sys.exit()


def __show_configs__(argv):
    parser.add_argument('configs', nargs='?', help='显示配置')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    from .config import list_server
    configs = list_server()
    out_print('---------- 配置信息 ----------')
    if not configs:
        out_print('             [空]')
        sys.exit()

    for server_name, apikey in configs.items():
        out_print(f'{server_name} {apikey}')

    sys.exit()


def get_nginx():
    from .nginx import Nginx
    return Nginx()


def __create_snapshot__(argv):
    parser.add_argument('backup', nargs='?', help='创建快照')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    out_print('---------- 创建快照 ----------')
    nginx = get_nginx()
    sid = nginx.create_snapshot()
    out_print(f'快照ID: {sid}')
    sys.exit()


def __preview_snapshot__(argv):
    parser.add_argument('preview', nargs='?', help='预览快照')
    parser.add_argument('sid', nargs='?', help='指定快照ID')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    if not args.sid:
        out_print('请输入快照ID')
        sys.exit()

    out_print('---------- 预览快照 ----------')
    nginx = get_nginx()
    configs = nginx.get_snapshot(args.sid)
    for filename, content in configs.items():
        out_print(f'# configuration file: {filename}')
        out_print(content)
    sys.exit()


def __apply_snapshot__(argv):
    parser.add_argument('apply', nargs='?', help='恢复快照')
    parser.add_argument('sid', nargs='?', help='指定快照ID')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    if not args.sid:
        out_print('请输入快照ID')
        sys.exit()

    out_print('---------- 恢复快照 ----------')
    if __confirm__(f'请确认是否恢复快照[{args.sid}]?'):
        nginx = get_nginx()
        nginx.apply_snapshot(args.sid)
    sys.exit()


def __delete_snapshot__(argv):
    parser.add_argument('delete', nargs='?', help='删除快照')
    parser.add_argument('sid', nargs='?', help='指定快照ID')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    if not args.sid:
        out_print('请输入快照ID')
        sys.exit()

    out_print('---------- 删除快照 ----------')
    if __confirm__(f'请确认是否删除快照[{args.sid}]?'):
        nginx = get_nginx()
        nginx.delete_snapshot(args.sid)
    sys.exit()


def __list_snapshot__(argv):
    parser.add_argument('list', nargs='?', help='列出快照')
    parser.add_argument('-h', '--help', action='store_true', help='显示帮助信息')

    args = parser.parse_args()

    if args.help:
        parser.print_help()
        sys.exit()

    nginx = get_nginx()
    out_print('---------- 快照列表 ----------')
    snapshots = nginx.list_snapshot()
    if not snapshots:
        out_print('          [没有快照]')
        out_print('请使用 [salesea -b] 创建快照')
        sys.exit()
    for snapshot in snapshots:
        out_print(snapshot.get('name'))
    sys.exit()



# -*- coding: utf-8 -*-
from io import open
from setuptools import setup, find_packages

# ------------------------------------------------------------------------------- #

with open('README.md', 'r', encoding='utf-8') as fp:
    readme = fp.read()

# with open("requirements.txt", "r", encoding="utf8") as f:
#     requires = f.read()
#     print(requires.splitlines())
# ------------------------------------------------------------------------------- #

setup(
    name='salesea',
    author='howard',
    author_email='18071131140telephone@gmail.com',
    version='1.0.34',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    description='This is an Nginx log collection tool.',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords=[
        'nginx',
        'logs',
        'collection'
    ],
    # package_data={'': ['*.*']},
    # install_requires=['aiohttp>=3.8.4',
    #                   'aiosignal>=1.2.0',
    #                   'async-timeout>=4.0.2',
    #                   'attrs>=21.4.0',
    #                   'charset-normalizer>=2.0.7',
    #                   'frozenlist>=1.2.0',
    #                   'idna>=3.3',
    #                   'multidict>=5.2.0',
    #                   'yarl>=1.7.2'],
    install_requires=[
        'requests>=2.18.4',
        ''
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    entry_points={'console_scripts': "salesea = salesea.__main__:main"},
    python_requires='>=3.0',
)

# ------------------------------------------------------------------------------- #



# coding: utf-8
import hashlib
import json
import re
import os
import platform
from datetime import datetime
from pathlib import Path
from typing import List

from .log import logger


class Server:
    def __init__(self, nginx: "Nginx", **kwargs):
        self.nginx = nginx
        self.names: tuple = kwargs.get("names", tuple())
        self.default_name = self.names[0] if self.names else None
        self.listens = kwargs.get("listens", None)
        self.locations = kwargs.get("locations", None)
        self.includes = kwargs.get("includes", None)
        self.serverConfigPath = kwargs.get("serverConfigPath", None)

        if platform.system() == 'Windows':
            self.log_path = f"./logs/salesea-{self.names[0]}.log"
            self.salesea_conf_name = f"{self.names[0]}.salesea.conf"
            self.salesea_conf_dir = Path(os.path.expandvars(r"%LOCALAPPDATA%")) / "salesea" / "nginx" / "conf"
            self.salesea_conf_path = self.salesea_conf_dir / self.salesea_conf_name
        else:
            self.log_path = f"/var/log/nginx/salesea-{self.names[0]}.log"
            self.salesea_conf_name = f"{self.names[0]}.salesea.conf"
            self.salesea_conf_dir = Path("/etc/salesea/nginx/conf/")
            self.salesea_conf_path = self.salesea_conf_dir / self.salesea_conf_name

        logfile = Path(self.log_path)
        self.logfile = logfile if logfile.is_absolute() else Path(self.nginx.nginx_path).parent / logfile
        self.logfile.parent.mkdir(parents=True, exist_ok=True)

        self.root = kwargs.get("root", None)
        if self.root is not None:
            root = Path(self.root)
            self.root = root if root.is_absolute() else Path(self.nginx.nginx_path).parent / root

    def eq_name(self, server_list):
        # pattern = server_name.replace(".", "\.").replace("*", ".*?") if server_name is not None else None
        # if server_name is None:
        #     return False
        # for _ in self.names:
        #     if re.match(pattern, _) is not None:
        #         return True
        #
        # return False
        if not server_list:
            return False

        for server_name in server_list:
            if server_name in self.names:
                self.default_name = server_name
                return True

        return False

    def eq_listen(self, listen):
        if not listen:
            return False
        return str(listen.strip()) in self.listens

    def is_salesea(self):
        for include in self.includes:
            if include.endswith(".salesea.conf"):
                return True

    # 读取配置文件, 报错返回None
    def read_config(self):
        path = Path(self.salesea_conf_path)
        if not path.exists():
            return None

        with path.open(mode="r", encoding="utf-8") as fobj:
            return fobj.read()

    def write_config(self, content):
        path = Path(self.salesea_conf_path)
        with path.open(mode="w", encoding="utf-8") as fobj:
            logger.debug(f"写入[{self.salesea_conf_path}]文件")
            fobj.write(content)

    def include_salesea(self, api_key):
        logger.debug(f"检查[{self.names[0]}]服务是否包含[salesea.conf]")
        global_flag = 0

        self.salesea_conf_dir.mkdir(parents=True, exist_ok=True)
        salesea_conf = f'access_log {self.log_path} salesea;\n\n'
        salesea_conf += 'if ($uri = /salesea) {\n'
        salesea_conf += f'    return 200 "{api_key}";\n'
        salesea_conf += '}\n'
        # salesea_conf += 'if ($uri = /salesea) {\n'
        # salesea_conf += '    rewrite ^(.*)$ $1 last;\n'
        # salesea_conf += '}\n\n'
        # salesea_conf += 'location = /salesea {\n'
        # salesea_conf += '    add_header Content-Type text/plain;\n'
        # salesea_conf += f'    return 200 "{api_key}";\n'
        # salesea_conf += '}'

        # 方案一 将salesea的APIKEY写入静态文件
        # if self.root is not None:
        #     self.root.mkdir(parents=True, exist_ok=True)
        #     salesea_path = self.root / 'salesea'
        #
        #     try:
        #         salesea_path.write_text(api_key, encoding='utf-8')
        #     except Exception as e:
        #         logger.error(f'写入文件失败: {e}')

        # 方案二 将salesea的APIKEY写入NGINX配置文件
        if self.read_config() != salesea_conf:
            global_flag = 1
            self.write_config(salesea_conf)

        if self.is_salesea():
            return global_flag

        content = self.nginx.nginxConfigs.get(self.serverConfigPath)
        server_name = r'\s+'.join(self.names)

        # 行数
        servers = []
        include_ln = -1
        listen_ln = -1
        server_name_ln = -1
        listens = []

        # 常量标识
        line_num = 0
        flag = False
        flag_name = False
        block = ''
        num_of_quote = 0

        # 去除注释 除了# configuration file
        delete_comment = lambda x: x.group(0) if x.group(0).startswith('# configuration file ') else ''
        content1 = re.sub(r'#[^\n]*', delete_comment, content)

        for line in content1.splitlines():
            line_num += 1
            x = line.replace(' ', '').replace('\t', '')
            _ = line.strip()

            # 查询是否包含 include
            result = re.search(r"^include\s+.*?" + self.salesea_conf_name + ";", _)
            if result is not None:
                include_ln = line_num

            # 查询是否包含 listen
            result = re.search(r"^listen\s*((?:\d|\s)*)[^;]*;", _)
            if result is not None:
                listens.append(result.group(1).strip())
                listen_ln = line_num

            # 查询是否包含 server_name
            result = re.search(rf"^server_name\s+{server_name}\s*;", _)
            if result is not None:
                server_name_ln = line_num

            if x == 'server':
                flag_name = True

            if x.startswith('server{') or (flag_name and x.startswith('{')):
                num_of_quote += 1
                flag = True
                block += line
                continue

            if flag and '{' in line:
                num_of_quote += 1

            if flag and num_of_quote != 0:
                block += line

            if flag and '}' in line:
                num_of_quote -= 1

            if flag and num_of_quote == 0:
                servers.append({
                    'include_ln': include_ln,
                    'listen_ln': listen_ln,
                    'server_name_ln': server_name_ln,
                    'listens': tuple(listens),
                })
                flag = False
                flag_name = False
                block = ''
                num_of_quote = 0
                include_ln = -1
                listen_ln = -1
                server_name_ln = -1
                listens = []

        for server in servers:
            if server['listens'] == self.listens and not server['server_name_ln'] == -1 and server['include_ln'] == -1:
                logger.info(f"更新[{self.serverConfigPath}]文件")
                content = content.splitlines()
                # 获取缩进
                salesea_conf_path = self.salesea_conf_path.__str__().replace("\\", "/")
                indent = re.search(r'^(\s*)', content[server['server_name_ln'] - 1]).group(1)
                include_line = f'{indent}# --------Salesea Config Start--------\n'
                include_line += f'{indent}include {salesea_conf_path};'
                include_line += f'\n{indent}# --------Salesea Config End--------'
                content.insert(server['server_name_ln'], include_line)
                content = '\n'.join(content)
                self.nginx.nginxConfigs[self.serverConfigPath] = content
                with open(self.serverConfigPath, mode="w", encoding="utf-8") as fobj:
                    fobj.write(content.strip('\n') + '\n')

                global_flag = 1
                break


        return global_flag

    def __repr__(self):
        return f"<Server {self.names}:{self.listens} {self.includes} {self.serverConfigPath}>"


class NginxException(Exception):
    pass


class Nginx:

    def __init__(self, nginx_path=None):
        self.nginx_conf = None
        self.nginx_path = nginx_path
        # 分文件的nginx配置
        self.nginxConfigs = None
        self.httpConfigPath = None
        self.backend = None  # 保存后端ip和pool name
        self.httpBlock = None  # 保存解析后端每个http块
        self.logFormats = None  # 保存解析后端每个log_format块
        self.includes = None  # 保存解析后端每个include块
        self.serverBlock = None  # 保存解析后端每个server块
        self.servers: List[Server] = None
        self.init()

    def init(self):
        self.nginxConfigs = dict()
        self.httpConfigPath = None
        self.backend = list()
        self.httpBlock = list()
        self.logFormats = dict()
        self.includes = list()
        self.serverBlock = list()
        self.servers = list()
        self.load_nginx_config()
        self.parse_file()
        self.parse_http_block()
        self.parse_server_block()

    def is_salesea(self):
        return self.logFormats.get("salesea") is not None

    def write_salesea_log_format(self):
        logger.debug("检查[log_format salesea]状态")
        if self.is_salesea():
            return
        salesea_log_format = """    # --------Salesea Config Start--------
    log_format salesea '$remote_addr - $remote_user [$time_local] "$request" '
             '$status $body_bytes_sent "$http_referer" '
             '"$http_user_agent" "$http_x_forwarded_for"';
    # --------Salesea Config End--------"""

        content = self.nginxConfigs.get(self.httpConfigPath)
        if re.search(r"(\s*log_format\s+salesea\s+)", content) is None:
            logger.info(f"更新[{self.httpConfigPath}]文件")
            # 在http块前不能有注释符号# 前面可能是\n
            content = re.sub(
                r"(\n[\s\t]*http\s+\{)",
                r"\1\n" + salesea_log_format,
                content
            )
            with open(self.httpConfigPath, "w") as fp:
                fp.write(content.strip('\n') + '\n')

            self.reload()
            self.init()

    def include_saleseas(self, server_map):
        logger.debug("检查[server_name.salesea.conf]状态")
        flag = 0
        server_list = tuple(server_map.keys())
        for server in self.servers:
            if server.eq_name(server_list) and server.eq_listen("443"):
                flag += server.include_salesea(server_map.get(server.default_name))

        if flag > 0:
            self.reload()
            self.init()

    def load_nginx_config(self):
        # 判断平台
        if platform.system() == 'Windows':
            self.get_nginx_path_win()
            cmd = f'cd /d "{Path(self.nginx_path).parent}" && nginx -T'
            with os.popen(cmd) as fp:
                bf = fp._stream.buffer.read()
                self.nginx_conf = bf.decode("utf-8", errors="ignore")
        else:
            self.get_nginx_path_linux()
            self.nginx_conf = os.popen(f"{self.nginx_path} -T").read()

    # 拆分nginx配置文件
    def parse_file(self):
        current_path = None
        for line in self.nginx_conf.splitlines():
            config_file = re.findall('^# configuration\s+file\s+(.*):', line)
            if len(config_file) > 0:
                current_path = config_file[0]
                self.nginxConfigs[current_path] = ""
            else:
                if current_path is not None:
                    self.nginxConfigs[current_path] += line + "\n"

    @property
    def snapshot_dir(self):
        if platform.system() == 'Windows':
            path = Path(os.path.expandvars(r"%LOCALAPPDATA%")) / "salesea" / "nginx" / "snapshot"
        else:
            path = Path("/etc/salesea/nginx/snapshot")

        path.mkdir(parents=True, exist_ok=True)

        return path

    # 存储快照
    def create_snapshot(self):
        content = json.dumps(self.nginxConfigs, indent=4, ensure_ascii=False)
        snapshot_name = datetime.now().strftime("%Y%m%d%H%M%S")
        hash_md5 = str(hashlib.md5(content.encode("utf-8")).hexdigest())
        filename = hash_md5 + '-' + snapshot_name
        snapshot_dir = self.snapshot_dir
        snapshot_save_path = snapshot_dir / f"{filename}.json"
        # 通过hash_md5判断是否已经存在
        search = list(snapshot_dir.glob(f"{hash_md5}-*.json"))
        if search:
            logger.debug(f"快照[{search[0].name}]已经存在")
            return search[0]

        with open(snapshot_save_path, "w", encoding="utf-8") as fp:
            logger.info(f"保存快照[{snapshot_save_path.name}]")
            fp.write(
                json.dumps({
                    "name": snapshot_save_path.name,
                    "type": "nginx",
                    "content": content,
                }, indent=4, ensure_ascii=False)
            )

        return filename

    # 获取快照列表
    def list_snapshot(self):
        return [
            {
                "name": f.name.replace(".json", ""),
                "mtime": f.stat().st_mtime,
                "size": f.stat().st_size
            }
            for f in self.snapshot_dir.glob("*-*.json")
        ]

    # 恢复快照
    def apply_snapshot(self, filename):
        snapshot_path = self.snapshot_dir / f'{filename}.json'
        if not snapshot_path.exists():
            logger.error(f"快照[{filename}]不存在")
            return False

        logger.info(f"恢复快照[{filename}]")
        with open(snapshot_path, "r", encoding="utf-8") as fp:
            content = fp.read()
            obj = json.loads(content)
            if obj.get("type") != "nginx" or obj.get("content") is None or obj.get("name") is None:
                logger.error(f"快照[{filename}]已损坏")
                return False
            nginxConfigs = json.loads(obj.get("content"))
            for path, content in nginxConfigs.items():
                with open(path, "w", encoding="utf-8") as fp:
                    fp.write(content)
        logger.info(f"恢复快照[{filename}]成功")
        self.reload()
        return True

    # 删除快照
    def delete_snapshot(self, filename):
        snapshot_path = self.snapshot_dir / f'{filename}.json'
        if not snapshot_path.exists():
            logger.error(f"快照[{filename}]不存在")
            return False

        logger.info(f"文件[{snapshot_path}]已删除")
        snapshot_path.unlink()
        return True

    # 预览快照
    def get_snapshot(self, filename):
        snapshot_path = self.snapshot_dir / f'{filename}.json'
        if not snapshot_path.exists():
            logger.error(f"快照[{filename}]不存在")
            return False

        with open(snapshot_path, "r", encoding="utf-8") as fp:
            content = fp.read()
            obj = json.loads(content)
            if obj.get("type") != "nginx" or obj.get("content") is None or obj.get("name") is None:
                logger.error(f"快照[{filename}]已损坏")
                return False
            nginxConfigs = json.loads(obj.get("content"))
            return nginxConfigs

    # nginx reload  重载nginx配置
    def reload(self):
        if platform.system() == 'Windows':
            cmd = f'cd /d "{Path(self.nginx_path).parent}" && nginx -s reload'
            os.popen(cmd).read()
        else:
            os.popen(f"{self.nginx_path} -s reload").read()

    #############################################################################
    #####Parse Nginx Config######################################################
    def get_nginx_path_win(self):
        os.popen(f"chcp 65001")
        if self.nginx_path is not None and self.nginx_path.exists():
            return self.nginx_path

        nginx_paths = os.popen(f"where nginx").readlines()

        if not nginx_paths:
            raise NginxException('请将[nginx.exe]配置到环境变量中')

        for path in nginx_paths:
            path = Path(path.strip())
            if path.exists():
                self.nginx_path = path
                return path

    def get_nginx_path_linux(self):
        if self.nginx_path is not None and self.nginx_path.exists():
            return self.nginx_path

        nginx_paths = os.popen(f"which nginx").readlines()

        if not nginx_paths:
            raise NginxException('请将[nginx]配置到环境变量中')

        for path in nginx_paths:
            path = Path(path.strip())
            if path.exists():
                self.nginx_path = path
                return path

    def parse_root(self, content):
        flag = False
        result = None
        num_of_quote = 0
        for line in content.splitlines():
            x = line.replace(' ', '').replace('\t', '')

            if x.startswith('root') and num_of_quote == 1:
                # 匹配line , 例如: root /usr/share/nginx/html; 可能有引号或者双引号
                result = re.search(r'root\s+["\']?([^"\']+)["\']?;', line).group(1)

            if x.startswith('{'):
                flag = True
                flag_name = False

            if flag and '{' in line:
                num_of_quote += 1

            if flag and '}' in line:
                num_of_quote -= 1

            if flag and num_of_quote == 0:
                return result

    def parse_block(self, block_name, block_content=None):
        flag = False
        flag_name = False
        block = ''
        file = ''
        result = []
        num_of_quote = 0

        content = (block_content or self.nginx_conf)

        # 去除注释 除了# configuration file
        delete_comment = lambda x: x.group(0) if x.group(0).startswith('# configuration file ') else ''
        content = re.sub(r'#[^\n]*', delete_comment, content)

        for line in content.splitlines():
            x = line.replace(' ', '').replace('\t', '')

            config_file = re.findall('^# configuration\s+file\s+(.*):', line)
            if len(config_file) > 0:
                file = config_file[0]

            if x == block_name:
                flag_name = True

            if x.startswith(block_name + '{') or (flag_name and x.startswith('{')):
                num_of_quote += 1
                flag = True
                block += line
                continue

            # 发现{，计数加1。发现}，计数减1，直到计数为0
            if flag and '{' in line:
                num_of_quote += 1

            # 将proxy_pass的别名换成ip
            # if flag and 'proxy_pass' in line:
            #     r = re.findall('proxy_pass\s+https?://([^;/]*)[^;]*;', line)
            #     if len(r) > 0:
            #         for pool in self.backend:
            #             if r[0] == pool['poolname']:
            #                 line = line.replace(r[0], pool['ip'])

            if flag and num_of_quote != 0:
                block += line + '\n'

            if flag and '}' in line:
                num_of_quote -= 1

            if flag and num_of_quote == 0:
                result.append({
                    'block': block,
                    'file': file
                })
                flag = False
                flag_name = False
                block = ''
                num_of_quote = 0

        return result

    def parse_server_block(self):
        self.serverBlock = self.parse_block('server')

        if len(self.serverBlock) == 0:
            raise NginxException('没有解析到[server]配置文件')

        for block in self.serverBlock:
            serverBlock = block['block']
            serverConfigPath = block['file']

            # listen有可能有多个
            listens = tuple([_.strip() for _ in re.findall('listen\s*((?:\d|\s)*)[^;]*;', serverBlock)])
            r = re.findall('server_name\s+([^;]*);', serverBlock)  # server_name只有一个

            # 可能存在没有server_name的情况
            if not len(r) > 0:
                continue

            server_names = tuple(r[0].split())

            # 判断servername是否有ip，有ip就不存。比如servername 127.0.0.1这样的配置
            for server_name in server_names:
                if len(re.findall('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', server_name)) > 0:
                    continue

            includes = [_.strip() for _ in re.findall('include\s+([^;]*);', serverBlock)]
            # location可能不止一个
            locations = re.findall('location\s*[\^~\*=]{0,3}\s*([^{ ]*)\s*\{[^}]*proxy_pass\s+https?://([^;/]*)[^;]*;',
                                   serverBlock)

            backend_list = list()
            backend_ip = ''

            # 可能存在多个location
            if len(locations) > 0:
                for location in locations:
                    backend_path = location[0]
                    poolname = location[1]
                    # 如果不是ip的pool name，就取出后端对应的ip
                    if len(re.findall('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', poolname)) == 0:
                        for backend in self.backend:
                            if poolname == backend['poolname']:
                                backend_ip = backend['ip']
                                break
                    else:
                        backend_ip = poolname

                    backend_list.append({"backend_path": backend_path, "backend_ip": backend_ip})

            # 获取root 在location外面的root
            root = self.parse_root(serverBlock)

            server = Server(
                self,
                **{
                    'names': server_names,
                    'listens': listens,
                    'root': root,
                    'locations': locations,
                    'includes': includes,
                    'backend': backend_list,
                    'serverConfigPath': serverConfigPath,
                }
            )

            self.servers.append(server)

    def parse_http_block(self):
        self.httpBlock = self.parse_block('http')
        httpConfigPath = None

        if len(self.httpBlock) == 0:
            raise NginxException('没有解析到[http]配置文件')

        for block in self.httpBlock:
            httpBlock = block['block']
            httpConfigPath = self.httpConfigPath or block['file']

            # 获取log_format log_format后面不能是\n
            for log_format in re.findall('\s*log_format\s+([^;]*);', httpBlock):
                log_format = log_format.strip()
                name, log_format = log_format.split(' ', 1)
                self.logFormats[name] = log_format

            # 获取include
            self.includes += [_.strip() for _ in re.findall('include\s+([^;]*);', httpBlock)]

            # 获取upstream
            self.upstreams = re.findall('upstream\s+([^{ ]+)\s*{([^}]*)}', httpBlock)

        self.httpConfigPath = httpConfigPath
